# -*- coding: utf-8 -*-
"""
Curate, visualize, annotate, and share your behavioral ephys data using Python
"""

from .version import version as __version__
from .version import git_revision as __git_revision__

from .datasets import *
from .gui import *
from .scripts import *
